/*-----------------------------------------------------------------------

      Copyright (c) 1998,1999 Eiwa System Management, Inc

                      All Rights Reserved

  This file contains proprietary source code and other proprietary
  material owned by Eiwa System Management, Inc. Published only in
  limited copyright sence.

------------------------------------------------------------------------*/
/*
 * $Id: TJEventSource.h,v 1.11 1999/07/28 06:01:58 hiranabe Exp $
 */
#if !defined(TJEventSource_h)
#define TJEventSource_h

#ifndef tjconfig_h
#include "tjconfig.h"
#endif

#include <vector>
#ifndef TJEventListener_h
#include "TJEventListener.h"
#endif
#ifndef tjstdnamespace_h
#include "tjstdnamespace.h"
#endif

/**
 * Cxg\[X.
 *
 * eCxg̃\[X()̎x邽߂̃ev[głC
 * Cxg^ƂĂDEventListener Ƒg Java ̃Cxgf
 * java.util.EventObject / java.util.EventListener T|[gD
 *
 * C++ ł͑dpp邱Ƃł邽߁CCxg\[X
 * ɂpłT|[gNXƂẴNXpӂD
 * (java.beans.PropertyChangesSupportQ)
 *
 * o^ꂽXiւ̒ʒm̏͋K肳ĂȂD
 *
 * o^ꂽ Listener  delete(fXgN^Ăяo) ɂẮC
 * {Iɂ̓AvP[V̐ӔCƂ邪CI destroyListeners 
 * ĂԂƂőS Listener ɑ΂ delete ĂяoƂłD
 *
 * Event, Listener, Source ̎gpD
 *
 * \begin{verbatim}
 *
 *   // AvP[V`̃Cxg^
 *   class XxxEvent {
 *        TJEventSource<XxxEvent>* source;
 *        int value;
 *   public:
 *        XxxEvent(TJEventSource<XxxEvent>* s, int v):
 *            source(s), value(v) { }
 *        TJEventSource<XxxEvent>* getSource() const {
 *             return source;
 *        }
 *        int getValue() const { return value; }
 *   };
 *
 *   // AvP[V`̃CxgXi
 *   #include "TJEventListener.h"
 *   class XxxListener : public TJEventListener<XxxEvent> {
 *       public:
 *           virtual void eventHappened(const XxxEvent& ev) {
 *             cout << "ev.value=" << ev.getValue() << endl;
 *           // }`Xbhł this IuWFNg̑͊댯
 *           }
 *   };
 *
 *   // AvP[V`̃Cxg\[X
 *   #include "TJEventSource.h"
 *   class XxxSource : public TJEventSource<XxxEvent> {
 *       public:
 *           void foo() {
 *                // ........
 *                // Cxg
 *                XxxEvent ev(this, 2);
 *                fireEvent(ev);  // eventHappened Ăяo
 *           }
 *   };
 *
 *   // \[XƃXi̘A
 *   XxxSource source;
 *   XxxListener listener;
 *   source.addListener(&listener);
 *
 * \end{verbatim}
 *
 * Listener, Source ̌Ăяo֌W͈ȉ̂悤ɂȂD
 * 
 * \begin{verbatim}
 *
 *        XxxSource                        XxxListener
 *   method() |                                 |
 * ---------> |                                 |
 *            |--+                              |
 *            |  |fireEvent()                   |
 *            |<-+                              |
 *            |                                 |
 *            |          eventHappend()         |
 *            |-------------------------------->|
 *            |                                 |
 *
 * \end{verbatim}
 *
 * [bNj]
 *
 * ̃NX́C}`XbhɊւ郍bNSӎȂD
 * Kvł TJEventSource D
 *
 * @author tokui, hiranabe
 * @date 1998,7/31
 * @date 1999,4/1
 * @see TJSafeEventSource
 * @see TJEventListener
 */
template<class EventType> class TJEventSource {
public:
	/**
	 * Xǐ^.
	 */
	typedef TJEventListener<EventType> ListenerType;
protected:
	/**
	 * Xi
	 */
	vector<ListenerType*> listeners;

public:
	/**
	 * fXgN^.
	 * Listener ̃fXgN^͌ĂяoȂƂɒӁD
	 *
	 * @see TJEventSource#destroyEventListeners()
	 */
	virtual ~TJEventSource() { }

	/**
	 * CxgXio^.
	 *
	 * @param l o^郊XiDnull Ȃ疳
	 */
	virtual void addEventListener(ListenerType* l) {
		if (l != 0)
			listeners.push_back(l);
	}

	/**
	 * CxgXi폜.
	 *
	 * @param l 폜郊XiDnull Ȃ疳
	 */
	virtual void removeEventListener(ListenerType* l) {
		if (l == 0)
			return;
		for (vector<ListenerType*>::iterator i = listeners.begin();
			i != listeners.end(); ++i) {
			if ((*i) == l) {
				listeners.erase(i);
				break;
			}
		}
	}

	/**
	 * o^Ă邷ׂĂ Listener 폜.
	 * ̃\bhĂ΂ꂽCʒmׂ Listener ̃Xg͋ɂȂD
	 * e Listener  delete (fXgN^Ăяo)͍sȂD
	 */
	virtual void removeEventListeners() {
		listeners.clear();
	}

	/**
	 * Listner Ă邩ǂ̌.
	 *
	 * @param l Listener IuWFNg
	 * @return Ă true
	 */
	virtual bool hasEventListener(const ListenerType* l) const {
		if (l == 0)
			return false;
		for (vector<ListenerType*>::const_iterator i = listeners.begin();
			i != listeners.end(); ++i) {
			if ((*i) == l)
				return true;
		}
		return false;
	}

	/**
	 * o^Ă Listener  delete (fXgN^Ăяo).
	 *
	 * {I Listener  delete ̓AvP[V̐ӔCƂ邪C
	 * AvP[V Listener ̊ǗȂꍇC̃\bh
	 *  delete xD
	 * Listener ϐ̏ꍇ́CRgpĂ͂ȂȂD
	 */
	virtual void destroyEventListeners() const {
		for (int i = listeners.size() - 1; i >= 0; i--)
			delete listeners[i];
	}

	/**
	 * ݂ Listener ̐Ԃ.
	 *
	 * @return ݂ Listener ̐
	 */
	virtual unsigned countEventListeners() const {
		return listeners.size();
	}

protected:
	/**
	 * CxgXiɃCxgʒmD
	 *
	 * @param ev XiɒʒmCxg
	 */
	virtual void fireEvent(const EventType& ev) {
		for (unsigned n = listeners.size(), i = 0; i < n; i++)
			listeners[i]->eventHappened(ev);
	}
};


#endif	/* !defined(TJEventSource_h) */
