/*-----------------------------------------------------------------------

      Copyright (c) 1998,1999 Eiwa System Management, Inc

                      All Rights Reserved

  This file contains proprietary source code and other proprietary
  material owned by Eiwa System Management, Inc. Published only in
  limited copyright sence.

------------------------------------------------------------------------*/
/*
 * $Id: TJObservable.cc,v 1.6 1999/07/28 06:02:07 hiranabe Exp $
 */
#include "TJObservable.h"

#include "TJObserver.h"
#include <algorithm>

TJObservable::TJObservable(): changed(false) {
}

void TJObservable::addObserver(TJObserver* o) {
    if (o == 0)
        return;

    ObserverVector::const_iterator i = find(observer.begin(), observer.end(), o);
    if (i == observer.end())
        observer.push_back(o);
}

void TJObservable::removeObserver(TJObserver* o) {
    if (o == 0)
        return;

    ObserverVector::iterator i = find(observer.begin(), observer.end(), o);
    if (i != observer.end())
        observer.erase(i);
}

void TJObservable::notifyObservers() {
    notifyObservers(0);
}

void TJObservable::notifyObservers(void* arg) {
    if (!hasChanged())
        return;
    clearChanged();

    for (unsigned n = observer.size(), i = 0; i < n; i++)
        observer[i]->update(*this, arg); // no null observer
}

void TJObservable::removeObservers() {
    observer.clear();
}

bool TJObservable::hasChanged() const {
    return changed;
}

unsigned TJObservable::countObservers() const {
    return observer.size();
}

void TJObservable::destroyObservers() const {
    for (int i = observer.size() - 1; i >= 0; i--)
        delete observer[i];
}

void TJObservable::setChanged() {
    changed = true;
}

void TJObservable::clearChanged() {
    changed = false;
}

TJObservable::~TJObservable() {
}

bool TJObservable::hasObserver(const TJObserver* o) const {
    if (o == 0)
        return false;

    ObserverVector::const_iterator i = find(observer.begin(), observer.end(), o);
    return i != observer.end();
}


bool TJObservable::isSafe() const {
    return true;
}
