/*-----------------------------------------------------------------------

      Copyright (c) 1998,1999 Eiwa System Management, Inc

                      All Rights Reserved

  This file contains proprietary source code and other proprietary
  material owned by Eiwa System Management, Inc. Published only in
  limited copyright sence.

------------------------------------------------------------------------*/
/*
 * $Id: TJQueue.cc,v 1.4 1999/07/28 06:02:12 hiranabe Exp $
 */
#include "TJQueue.h"

#include <stdio.h>
#include "TJPrimitive.h"
#include "TJSystemInternal.h"
#include <limits.h>


const long TJQueue::INFINITE = (long)(-1);
const long TJQueue::NONBLOCK = (long)0;
const int  TJQueue::NODATA = INT_MAX;

unsigned TJQueue::serialno = 0;
static const char basename[] = "_mque_";


TJQueue::TJQueue(int maxmsgs, int maxmsglen) {
	if (maxmsgs <= 0 || maxmsglen <= 0)
		TJSystem_invalid("TJQueue::TJQueue: invalid args");
	maxlen = maxmsglen;
	::sprintf(name, "%s%d", basename, serialno++);
	int error = ::TJQueueCreate(name, maxmsgs, maxmsglen, &msgqueue);
	if (error != 0)
		TJSystem_Error("TJQueue::TJQueue", error);
}

TJQueue::~TJQueue() {
	::TJQueueDelete(msgqueue, name);
}

void TJQueue::send(const void* msg, int length) {
	if (length <= 0 || length > maxlen)
		TJSystem_invalid("TJQueue::send: invalid length");
	int error = ::TJQueueSend(msgqueue, msg, length);
	if (error == -1)
		TJSystem_invalid("TJQueue::send: deleted TJQueue");
	if (error != 0)
		TJSystem_Error("TJQueue::send", error);
}

int TJQueue::receive(void* msgbuf, int bufsize,
		long milliseconds /*=INFINITE*/) {
	int len = 0;
	/*
	 * TJQueueReceive() ̏o͏
	 *   milliseconds > 0 Ń^CAEgꍇ
	 *   milliseconds = 0  Mf[^݂Ȃꍇ
	 *    len = -1 Ŗ߂l 0 ł.
	 */
	int error = ::TJQueueReceive(msgqueue, msgbuf, bufsize, &len,
			milliseconds);
	if (error == -1)
		return 0;	// destroyed queue
	if (error != 0)
		TJSystem_Error("TJQueue::receive", error);
	return (len == -1)? NODATA : len;
}
