/*-----------------------------------------------------------------------

      Copyright (c) 1998,1999 Eiwa System Management, Inc

                      All Rights Reserved

  This file contains proprietary source code and other proprietary
  material owned by Eiwa System Management, Inc. Published only in
  limited copyright sence.

------------------------------------------------------------------------*/
/*
 * $Id: TJRWLock.cc,v 1.5 1999/04/14 02:43:17 tokui Exp $
 */
#include "TJRWLock.h"
#include "tjdebug.h"

#if defined(TJ_THREAD_DEBUG)
#	include "TJThreadDebugger.h"
#else
#	define TH_PUSHNAME(s)
#	define TH_POPNAME()
#endif


const long TJRWLock::INFINITE = (long)(-1);
const long TJRWLock::NONBLOCK = (long)0;

TJRWLock::TJRWLock() : rwlockMutex(),
		rlockSync(&rwlockMutex), wlockSync(&rwlockMutex) {
	rlockCount = 0;
	rlockWant = 0;
	wlockOwner = TJThread::None;
	wlockNest = 0;
	wlockWant = 0;
}

TJRWLock::~TJRWLock() {
}

/**
 * QƂ܂͍XV[hŃbNlD
 */
bool TJRWLock::beginTransaction(LockMode mode, long millis) {
	TH_PUSHNAME("TJRWLock::beginTransaction");
	bool signaled = true;
	rwlockMutex.lock();

	if (mode == EXCLUSIVE) {	// XV[h̃bNl
		TJThread::ID curtid = TJThread::currentThreadID();
		if (wlockOwner == curtid) {
			/*
			 * JgXbhXVbN
			 */
			wlockNest++;
		} else {
			while (wlockOwner != TJThread::None || rlockCount > 0) {
				if (millis == NONBLOCK) {
					rwlockMutex.unlock();
					TH_POPNAME();
					return false;	// cannot lock
				}
				wlockWant++;
				if (millis == INFINITE)
					wlockSync.wait();
				else
					signaled = wlockSync.timedwait(millis);
				wlockWant--;
			}
			/*
			 * XVbNԂłȂQƃbNԂłȂ̂
			 * XVbN
			 */
			wlockOwner = curtid;
			wlockNest = 1;
		}
	} else {	// Qƃ[h̃bNl
		while (wlockOwner != TJThread::None || wlockWant > 0) {
			if (millis == NONBLOCK) {
				rwlockMutex.unlock();
				TH_POPNAME();
				return false;	// cannot lock
			}
			rlockWant++;
			if (millis == INFINITE)
				rlockSync.wait();
			else
				signaled = rlockSync.timedwait(millis);
			rlockWant--;
		}
		/*
		 * XVbNԂłȂXVbNvȂ̂
		 * QƃbN
		 */
		rlockCount++;
	}

	rwlockMutex.unlock();
	TH_POPNAME();
	return signaled;
}

/**
 * bND
 *
 * @see H}`XbhvO~O (AXL[oŋ)
 */
void TJRWLock::endTransaction() {
	TH_PUSHNAME("TJRWLock::endTransaction");
	rwlockMutex.lock();

	if (wlockOwner != TJThread::None) {
		assert(wlockOwner == TJThread::currentThreadID());
		wlockNest--;
		assert(wlockNest >= 0);

		if (wlockNest == 0) {
			/*
			 * XVbN
			 */
			wlockOwner = TJThread::None;
		}
	} else {
		if (rlockCount > 0) {
			/*
			 * QƃbN
			 */
			rlockCount--;
		}
	}

	/*
	 * ̏̈Ӗ writer ҂ĂčXVbNl\
	 */
	bool writer = wlockWant > 0
		&& wlockOwner == TJThread::None && rlockCount <= 0;
	/*
	 * ̏̈Ӗ reader ҂ĂĎQƃbNl\
	 */
	bool reader = rlockWant > 0
		&& wlockOwner == TJThread::None && wlockWant <= 0;

	/*
	 * signal() ̑O unlock() D
	 * ̗R͑҂ĂXbh̃vCIeB signal()
	 * AOɑ҂ĂXbh𓾂ăbNꂸɖĂ
	 * ܂߂łD̂Ǝ̂͑botԂʂɂȂ邾
	 * CRxȏ̃vCIeBĂꍇ͂߂ǂ
	 * ɂȂ\傫ȂD
	 * ylz̑΍ŃvCIeB̖肪ȂȂ邱Ƃ͂ȂD
	 */
	rwlockMutex.unlock();

	/*
	 * writer Dő҂ĂXbhfBԂɂD
	 */
	if (writer)
		wlockSync.signal();
	else if (reader)
		rlockSync.signal();

	TH_POPNAME();
}

bool TJRWLock::isWLocked() const {
	return wlockOwner != TJThread::None;	// atomic operation
}

bool TJRWLock::isRLocked() const {
	return rlockCount > 0;	// atomic operation
}

bool TJRWLock::invariant() const {
	assert(wlockOwner == TJThread::None || rlockCount == 0);
	assert(rlockCount >= 0);
	assert(rlockWant >= 0);
	assert(wlockNest >= 0);
	assert(wlockWant >= 0);
	return true;
}
