/*-----------------------------------------------------------------------

      Copyright (c) 1998,1999 Eiwa System Management, Inc

                      All Rights Reserved

  This file contains proprietary source code and other proprietary
  material owned by Eiwa System Management, Inc. Published only in
  limited copyright sence.

------------------------------------------------------------------------*/
/*
 * $Id: TJSafeEventSource.h,v 1.12 1999/07/28 06:02:20 hiranabe Exp $
 */
#if !defined(TJSafeEventSource_h)
#define TJSafeEventSource_h

#ifndef tjconfig_h
#include "tjconfig.h"
#endif

#include <vector>
#ifndef TJSafeEventListener_h
#include "TJSafeEventListener.h"
#endif
#ifndef TJEventSource_h
#include "TJEventSource.h"
#endif
#ifndef tjstdnamespace_h
#include "tjstdnamespace.h"
#endif

/**
 * XbhZ[tȃCxg\[X.
 *
 * eCxg̃\[X()̎x邽߂̃ev[gD
 *
 * TJSafeEventSource  TJSafeEventListener ɂCxg`B@\
 * ̓}`Xbhɂ闘pz肵̈S߂ĂD
 *
 * Event, SafeListener, SafeSource ̎gpD
 *
 * \begin{verbatim}
 *
 *  // AvP[V`̃Cxg^
 *  class XxxEvent {
 *       TJSafeEventSource<XxxEvent>* source;
 *       int value;
 *  public:
 *       XxxEvent(TJSafeEventSource<XxxEvent>* s, int v):
 *          source(s), value(v) {  }
 *       TJSafeEventSource<XxxEvent>* getSource() const {
 *          return source;
 *       }
 *       int getValue() const { return value; }
 *  };
 *
 *  // AvP[V`̃CxgXi
 *  #include "TJSafeEventListener.h"
 *   class XxxListener : public TJSafeEventListener<XxxEvent> {
 *   public:
 *       virtual void eventHappened(const XxxEvent& ev) {
 *         TJSafeEventSource<XxxEvent>* source = ev.getSource();
 *         cout << "ev.value=" << ev.getValue() << endl;
 *         // this IuWFNg͈̑S
 *         // ....
 *        }
 *   };
 *   // AvP[V`̃Cxg\[X
 *   #include "TJEventSource.h"
 *   class XxxSource : public TJSafeEventSource<XxxEvent> {
 *       public:
 *          void foo() {
 *             // ........
 *             // Cxg
 *             XxxEvent ev(this, 2);
 *             fireEvent(ev);  // safeEventHappened Ăяo
 *                             // safeEventHappened  eventHappened Ă
 *          }
 *   };
 *
 *   // \[XƃXi̘A
 *   XxxSource source;
 *   XxxListener listener;
 *   source.addListener(&listener);
 *
 * \end{verbatim}
 *
 * Listener, Source ̌Ăяo֌W͈ȉ̂悤ɂȂD
 *
 * \begin{verbatim}
 *
 *        XxxSource                      XxxListener
 *   method() |                                 |
 * ---------> |                                 |
 *            |--+                              |
 *            |  |fireEvent()                   |
 *            |<-+                              |
 *            |                                 |
 *            |ňU                     |
 *            |SbN                     |
 *            |          safeEventHappend()     |
 *            |-------------------------------->|ŃbN
 *            |                                 |擾D
 *            |                                 |--+
 *            |                                 |  | eventHappend()
 *            |          method()               |<-+
 *            |<--------------------------------|
 *
 * \end{verbatim}
 *
 * [bNj]
 *
 * ʂɁCSafeEventSource p XxxSource  SafeEventListenerp
 * XxxListener Ԃł́CXxxListener -> XxxSource Ƃ
 * ̌ĂяoRɋN\DCeventHappened()
 * ̌Ăяo XxxSource  XxxListenerւ̃R[obNƂȂD 
 *
 * AvP[V́C}`Xbh̃fbhbN邽߂
 * ̌邱ƁD
 *
 *   (1) \[X烊Xiւ̃Cxgʒmł̓\[X͈؂̃bN
 *       炸ɍsD
 *
 *   (2) Xi\[Xɑ΂Ăяo̓\[XŎIɃbND
 *
 * ̃NX̃Cxgʒm fireEvent() ͏L(1)̌ɏ]
 * DɏĽ(2)邽߃TuNX̃\bh
 * j^Iy[V(synchronized)Ƃꍇ,
 *
 * \begin{verbatim}
 *
 *     int XxxSource::method() {
 *             TJMonitor synchro__(this);  // O synchro__ ͔C
 *               ...
 *     }
 *
 * \end{verbatim}
 *
 * Ƃ΂悢D
 *
 * yfireEvent()  fireEventRetainingLock() ̈Ⴂz
 *
 *   (1) fireEvent() ̏ꍇ
 *
 *     ̃Cxg\[XbNĂΉăCxgʒm
 *     sIɎIɃbN𕜌D
 *
 *     _
 *
 *         ʓIȃbNKwƂ͋t̃bNlȂ߃fbh
 *         bN̊댯͏ȂȂD
 *
 *     _
 *
 *         bN fireEvent() ďoƂƈꎞIɃbN
 *         邽 fireEvent() 畜ÃCxg\[
 *         X̏ԂύXĂ\D
 *         ꂪssł fireEvent() Iтꍇ fireEvent()
 *         ďoOɃXbhɃRs[Ă@D
 *
 *   (2) fireEventRetainingLock() ̏ꍇ
 *
 *     ̃Cxg\[XbNĂꍇłɃCxg
 *     ʒmsD
 *
 *     _
 *
 *         ̃Cxg\[XbNĂꍇ fireEventRetainingLock()
 *         ďȏOŃ\[XIuWFNg̏Ԃ͕ۑD
 *
 *     _
 *
 *         ʓIȃbNKwƂ͋t̃bNl\
 *         fbhbN̊댯傫D
 *         fireEventRetainingLock() ͓IɊm̃R[hďoꍇ
 *         fbhbN̊댯ȂƂۏ؂łꍇ݂̂
 *         ׂłD
 */
template<class EventType> class TJSafeEventSource
    : public TJEventSource<EventType>, virtual public TJSyncObject {
public:
        /**
         * XbhZ[tȃXi^.
         */
        typedef TJSafeEventListener<EventType> SafeListenerType;

        /**
         * fXgN^.
         * Listener ̃fXgN^͌ĂяoȂƂɒӁD
         *
         * @see TJSafeEventSource#destroyObservers()
         */
        virtual ~TJSafeEventSource() { }

        /**
         * CxgXio^.
         *
         * @param l o^郊XiDnull Ȃ疳
         */
        virtual void addEventListener(ListenerType* l) {
            TJMonitor synchro__(this);
            TJEventSource<EventType>::addEventListener(l);
        }

        /**
         * CxgXi폜D
         *
         * @param l 폜郊XiDnull Ȃ疳
         */
        virtual void removeEventListener(ListenerType* l) {
            TJMonitor synchro__(this);
            TJEventSource<EventType>::removeEventListener(l);
        }

        /**
         * o^Ă邷ׂĂ Listener 폜.
         * ̃\bhĂ΂ꂽCʒmׂ Listener ̃Xg͋ɂȂD
         * e Listener  delete (fXgN^Ăяo)͍sȂD
         */
        virtual void removeEventListeners() {
            TJMonitor synchro__(this);
            TJEventSource<EventType>::removeEventListeners();
        }

        /**
         * Listner Ă邩ǂ̌.
         *
         * @param l Listener IuWFNg
         */
        virtual bool hasEventListener(const ListenerType* l) const {
            TJMonitor synchro__(this);
            return TJEventSource<EventType>::hasEventListener(l);
        }

        /**
         * o^Ă Listener  delete (fXgN^Ăяo).
         *
         * {I Listener  delete ̓AvP[V̐ӔCƂ邪C
         * AvP[V Listener ̊ǗȂꍇC̃\bh
         *  delete xD
         * Listener ϐ̏ꍇ́CRgpĂ͂ȂȂD
         */
        virtual void destroyEventListeners() const {
            TJMonitor synchro__(this);
            TJEventSource<EventType>::destroyEventListeners();
        }

        /**
         * ݂ Listener ̐Ԃ.
         *
         * @return ݂ Listener ̐
         */
        virtual unsigned countEventListeners() const {
            TJMonitor synchro__(this);
            return TJEventSource<EventType>::countEventListeners();
        }

protected:
        /**
         * CxgXiɃCxgʒm.
         *
         *yfireEvent()  fireEventRetainingLock() ̈ႢzQƂ̂ƁD
         * @param ev ʒmCxgIuWFNg
         */
        virtual void fireEvent(const EventType& ev) {
            if (countEventListeners() <= 0)
                return;

            vector<ListenerType*> vec;
            synchronized(this) {
                /*
                 * bNl listeners Rs[D
                 */
                vec = listeners;
                /*
                 * UbN
                 */
            }

            /*
             * SIuWFNgbNăCxgʒmsD
             */
            unsynchronized(this) {
                for (unsigned n = vec.size(), i = 0; i < n; i++) {
                    if (vec[i]->isSafe())
                        ((SafeListenerType*)vec[i])->safeEventHappened(ev);
                    else
                        vec[i]->eventHappened(ev);
                }

            }
        }

        /**
         * CxgXiɃCxgʒm.
         *
         *yfireEvent()  fireEventRetainingLock() ̈ႢzQƂ̂ƁD
         *
         * @param ev ʒmCxgIuWFNg
         */
        virtual void fireEventRetainingLock(const EventType& ev) {
            if (countEventListeners() <= 0)
                return;

            vector<ListenerType*> vec;
            synchronized(this) {
                /*
                 * bNl listeners Rs[D
                 */
                vec = listeners;
                /*
                 * UbN
                 */
            }

            /*
             * bNԂ͕ύXɃCxgʒmsD
             */
            for (unsigned n = vec.size(), i = 0; i < n; i++) {
                if (vec[i]->isSafe())
                    ((SafeListenerType*)vec[i])->safeEventHappened(ev);
                else
                    vec[i]->eventHappened(ev);
            }
        }

};


#endif	/* !defined(TJSafeEventSource_h) */
