/*-----------------------------------------------------------------------

      Copyright (c) 1998,1999 Eiwa System Management, Inc

                      All Rights Reserved

  This file contains proprietary source code and other proprietary
  material owned by Eiwa System Management, Inc. Published only in
  limited copyright sence.

------------------------------------------------------------------------*/
/*
 * $Id: TJSafeObservable.cc,v 1.6 1999/04/17 08:13:34 hiranabe Exp $
 */
#include "TJSafeObservable.h"
#include "TJSafeObserver.h"
#include "TJMonitor.h"
#include <algorithm>

TJSafeObservable::TJSafeObservable() {
}

void TJSafeObservable::addObserver(TJObserver* o) {
    synchronized(this) {
        TJObservable::addObserver(o);
    }
}

void TJSafeObservable::removeObserver(TJObserver* o) {
    synchronized(this) {
        TJObservable::removeObserver(o);
    }
}

void TJSafeObservable::notifyObservers() {
    notifyObservers(0);
}

void TJSafeObservable::notifyObservers(void* arg) {
    ObserverVector tmpObserver;

    synchronized(this) {
        /*
         * UbNāCupdate ĂяoɉD
         * PȃfbhbN̉\D
         */
        if (!hasChanged())
            return;
        tmpObserver = observer; // SvfRs[
        clearChanged();
    }

    /*
     * bNăCxgʒmsD
     * ̃\bhCbN擾̑̃\bhĂ΂ĂC
     * bNꎞIɊOD
     */
    unsynchronized(this) {
        for (int i = tmpObserver.size() - 1; i >= 0; i--) {
            TJObserver* o = tmpObserver[i];
            /*
             *  SafeObserver ȂCsafeUpdate Ăяo
             */
            if (o->isSafe()) {
                static_cast<TJSafeObserver*>(o)->safeUpdate(*this, arg);
            } else {
                o->update(*this, arg);
            }
        }
    }

}

void TJSafeObservable::notifyObserversRetainingLock() {
    notifyObserversRetainingLock(0);
}

void TJSafeObservable::notifyObserversRetainingLock(void* arg) {
    ObserverVector tmpObserver;

    synchronized(this) {
        /*
         * UbNāCupdate ĂяoɉD
         * PȃfbhbN̉\D
         */
        if (!hasChanged())
            return;
        tmpObserver = observer; // SvfRs[
        clearChanged();
    }
    /*
     * ݂̃bNԂۂ update ĂяoD
     * ̃\bhł̓bN͊OĂ悤ɂ݂邪C
     * ̃NXpNX synchronized \bh
     *  notify sꍇɁC̏ԂۂĂD
     */
    for (int i = tmpObserver.size() - 1; i >= 0; i--) {
        TJObserver* o = tmpObserver[i];
        /*
         *  SafeObserver ȂCsafeUpdate Ăяo
         */
        if (o->isSafe()) {
            static_cast<TJSafeObserver*>(o)->safeUpdate(*this, arg);
        } else {
            o->update(*this, arg);
        }
    }
}

void TJSafeObservable::removeObservers() {
    synchronized(this) {
        TJObservable::removeObservers();
    }
}

bool TJSafeObservable::hasChanged() const {
    return changed;  // atomic
}

unsigned TJSafeObservable::countObservers() const {
    TJMonitor sync(this);
    return TJObservable::countObservers();
}

void TJSafeObservable::destroyObservers() const {
    synchronized(this) {
        TJObservable::destroyObservers();
    }
}

void TJSafeObservable::setChanged() {
    synchronized(this) {
        TJObservable::setChanged();
    }
}

void TJSafeObservable::clearChanged() {
    synchronized(this) {
        TJObservable::clearChanged();
    }
}

TJSafeObservable::~TJSafeObservable() {
}

bool TJSafeObservable::hasObserver(const TJObserver* o) const {
    TJMonitor sync(this);
    return TJObservable::hasObserver(o);
}

bool TJSafeObservable::isSafe() const {
    return true;
}
