/*-----------------------------------------------------------------------

      Copyright (c) 1998,1999 Eiwa System Management, Inc

                      All Rights Reserved

  This file contains proprietary source code and other proprietary
  material owned by Eiwa System Management, Inc. Published only in
  limited copyright sence.

------------------------------------------------------------------------*/
/*
 * $Id: TJSafeObservable.h,v 1.7 1999/07/28 06:02:22 hiranabe Exp $
 */
#if !defined(TJSafeObservable_h)
#define TJSafeObservable_h

#ifndef tjconfig_h
#include "tjconfig.h"
#endif

#ifndef TJSyncObject_h
#include "TJSyncObject.h"
#endif
#ifndef TJObservable_h
#include "TJObservable.h"
#endif

/**
 * XbhZ[t Observable.
 *
 * [bNj]
 * ʂɁCObserver p ConcreteObserver ƁCObservable p
 * ConcreteObservable Ԃł́CConcreteObsever -> ConcreteObservable Ƃ
 * ̌ĂяoRɋNDCupdate() ̌ĂяóC
 * ConcreteObservable  ConcreteObserver ւ̃R[obNƂȂD 
 *
 * \begin{verbatim}
 *
 *      ConcreteObservable               ConcreteObserver
 *            |                                 |
 *  setState()|                                 |
 * ---------> |                                 |
 *            |--+                              |
 *            |  |setChanged()                  |
 *            |<-+                              |
 *  notify    |                                 |
 * Observers()|                                 |
 * ---------> |                                 |
 *            |ňU                     |
 *            |SbN                     |
 *            |          safeUpdate()           |
 *            |-------------------------------->|ŃbN
 *            |                                 |擾D
 *            |                                 |--+
 *            |                                 |  | update()
 *            |          getState()             |<-+
 *            |<--------------------------------|
 *
 * \end{verbatim}
 *
 * notifyObservers() ł́Cupdate() ̑OɌݎ擾Ă؂
 * bN邱ƂɂDɂāCfbhbN̉\
 * DAvP[V́CConcreteObserver -> ConcreteObservable 
 * Ăяo𐳏ƂāCbN̎擾̏Ƃ邱Ƃ]܂D
 *
 * @author hiranabe
 * @date 1999,3/30
 * @see TJSafeObserver
 * @see TJObservable
 * @see TJSafeEventSource
 */
class TJSafeObservable: public TJObservable, public virtual TJSyncObject {
public:
    /**
     * RXgN^.
     */
    TJSafeObservable();

    /**
     * Observer ǉ.
     *
     * @param o Observer IuWFNgDnull Ȃ疳D
     */
    virtual void addObserver(TJObserver* o);

    /**
     * Observer 폜.
     *
     * @param o Observer IuWFNg
     */
    virtual void removeObserver(TJObserver* o);

    /**
     * o^Ă邷ׂĂ Observer 폜.
     */
    virtual void removeObservers();

    /**
     * Observable ̕ω Observer ɒʒm.
     *
     * Observer  update ĂяȏOɁCbN
     * 邱ƂɒӁD
     */
    virtual void notifyObservers();

    /**
     * Observable ̕ω Observer ɒʒm.
     *
     * Observer  update ĂяȏOɁCbN
     * 邱ƂɒӁD
     *
     * @param arg AvP[VŎgpCӂ̃f[^
     */
    virtual void notifyObservers(void* arg);

    /**
     * Observable ̕ω Observer (bNێ̂܂)ʒm.
     *
     * Observer  update ĂяȏOɁCbN
     * sȂ߁CfbhbN̉\܂ނƂɒӁD
     */
    virtual void notifyObserversRetainingLock();

    /**
     * Observable ̕ω Observer (bNێ̂܂)ʒm.
     *
     * Observer  update ĂяȏOɁCbN
     * sȂ߁CfbhbN̉\܂ނƂɒӁD
     *
     * @param arg AvP[VŎgpCӂ̃f[^
     */
    virtual void notifyObserversRetainingLock(void* arg);

    /**
     * ωǂ.
     *
     * @return  true setChanged Ă΂܂ʒmĂȂ
     */
    virtual bool hasChanged() const;

    /**
     * ݂ Observer ̐Ԃ.
     *
     * @return ݂ Observer ̐
     */
    virtual unsigned countObservers() const;

    /**
     * o^Ă Observer  delete (fXgN^Ăяo).
     *
     */
    virtual void destroyObservers() const;

    /**
     * XbhZ[tȂ̂ǂ̌.
     *
     * ̃NXł́Ctrue ƂȂD
     * @return true XbhZ[t
     */
    virtual bool isSafe() const;

    /**
     * Observer Ă邩ǂ̌.
     *
     * @param o Oberver IuWFNg
     */
    virtual bool hasObserver(const TJObserver* o) const;

    /**
     * fXgN^.
     *
     * Observer ̃fXgN^͌ĂяoȂƂɒӁD
     * @see TJSafeObservable#destroyObservers()
     */
    virtual ~TJSafeObservable();

protected:
    /**
     * ωƂZbg.
     */
    virtual void setChanged();

    /**
     * ωƂNA.
     */
    virtual void clearChanged();
};

#endif	/* !defined(TJObservable_h) */
