/*-----------------------------------------------------------------------

      Copyright (c) 1998,1999 Eiwa System Management, Inc

                      All Rights Reserved

  This file contains proprietary source code and other proprietary
  material owned by Eiwa System Management, Inc. Published only in
  limited copyright sence.

------------------------------------------------------------------------*/
/*
 * $Id: TJSync.cc,v 1.8 1999/07/28 06:02:28 hiranabe Exp $
 */
#include "TJSync.h"

#include "TJMutex.h"
#include "TJPrimitive.h"
#include "TJSystemInternal.h"

#if defined(TJ_THREAD_DEBUG)
#	include "TJThreadDebugger.h"
#else
#	define TH_PUSHNAME(s)
#	define TH_POPNAME()
#	define TH_PAUSE(s)
#endif


TJSync::TJSync(const TJMutex* m) {
	int error = ::TJSyncCreate(&sync);
	if (error != 0)
		TJSystem_Error("TJSync::TJSync", error);
	mutex = (m != 0)? m->mutex : 0;
}

TJSync::~TJSync() {
	::TJSyncDelete(sync);
}

void TJSync::wait() {
	TH_PAUSE("before TJSync::wait");
	int error = ::TJSyncWait(sync, mutex);
	if (error != 0)
		TJSystem_Error("TJSync::wait", error);
	TH_PAUSE("TJSync::wait returned");
}

bool TJSync::timedwait(long milliseconds) {
	TH_PAUSE("before TJSync::timedwait");
	int error = ::TJSyncWaitTimeout(sync, mutex, milliseconds);
	if (error > 0)
		TJSystem_Error("TJSync::timedwait", error);
	bool signaled = error == 0;
	TH_PAUSE("TJSync::timedwait returned");
	return signaled;
}

void TJSync::signal() {
	TH_PAUSE("before TJSync::signal");
	int error = ::TJSyncNotify(sync);
	if (error != 0)
		TJSystem_Error("TJSync::signal", error);
	TH_PAUSE("TJSync::signal OK");
}

void TJSync::broadcast() {
	TH_PAUSE("before TJSync::broadcast");
	int error = ::TJSyncNotifyAll(sync);
	if (error != 0)
		TJSystem_Error("TJSync::broadcast", error);
	TH_PAUSE("TJSync::broadcast OK");
}
