/*-----------------------------------------------------------------------

      Copyright (c) 1998,1999 Eiwa System Management, Inc

                      All Rights Reserved

  This file contains proprietary source code and other proprietary
  material owned by Eiwa System Management, Inc. Published only in
  limited copyright sence.

------------------------------------------------------------------------*/
/*
 * $Id: TJSysError.c,v 1.8 1999/07/19 00:14:18 tokui Exp $
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "TJSysError.h"
#include "TJPrimitive.h"
#include "tjprimdebug.h"


#define MBUFSIZE 128

/**
 * Õnh
 */
static void (*errorHandler)() = 0;

/**
 * bZ[Wo͗p̊֐
 */
static void (*outputHandler)(const char*) = 0;


/**
 * bZ[Wo͂D
 * ftHgo͐ stderr łD
 */
static void writeMessage(const char* msg) {
	if (outputHandler != 0) {
		(*outputHandler)(msg);
	} else {
		fputs(msg, stderr);
		fputs("\n", stderr);
	}
}

/**
 * vIG[ɂ菈~D
 * ftHg̏̓bZ[Wo͂ abort() ĂяoD
 */
static void halt(const char* msg) {
	TJTRACE_GET();
	writeMessage(msg);
	if (errorHandler != 0)
		(*errorHandler)();
	abort();
}

/**
 * 烁bZ[WD
 */
static void compose(char* bufp, int bufsize, const char* id, const char* file,
		int line, const char* msg, int error) {

	static char sep1[] = " ", sep2[] = ":", sep3[] = " ";
	static char errinfo[] = " errno=";
	int len = 0;
	const char* errmsg;
	char lno[16];

	sprintf(lno, "%d", line);
	*bufp = '\0';
	if ((len += strlen(id) + strlen(sep1)) < bufsize) {
		strcat(bufp, id);
		strcat(bufp, sep1);
	}
	if ((len += strlen(file) + strlen(sep2)) < bufsize) {
		strcat(bufp, file);
		strcat(bufp, sep2);
	}
	if ((len += strlen(lno) + strlen(sep3)) < bufsize) {
		strcat(bufp, lno);
		strcat(bufp, sep3);
	}
	if ((len += strlen(msg)) < bufsize) {
		strcat(bufp, msg);
	}
	if (error != 0) {
		errmsg = TJGetErrorMessage(error);
		if ((len += strlen(errinfo) + strlen(errmsg)) < bufsize) {
			strcat(bufp, errinfo);
			strcat(bufp, errmsg);
		}
	}
}

/* -------------------------- public method ------------------------------- */

/**
 * Ôbnhݒ肷D
 * ݒȂ abort() ̌ĂяoƂȂD
 */
void TJSetErrorHandler(void (*hndlr)(void)) {
	errorHandler = hndlr;
}

/**
 * bZ[Wo͗p̂b֐w肷D
 * wȂ stderr ɏo͂D
 */
void TJSetOutputHandler(void (*out)(const char*)) {
	outputHandler = out;
}

void TJNomemPanic(const char* file, int line, const char* msg) {
	char buf[MBUFSIZE];
	compose(buf, sizeof(buf), "nomem!!", file, line, msg, 0);
	halt(buf);
}

void TJInconsistentPanic(const char* file, int line,
		const char* msg) {
	char buf[MBUFSIZE];
	compose(buf, sizeof(buf), "inconsistent!", file, line, msg, 0);
	halt(buf);
}

void TJInvalidPanic(const char* file, int line, const char* msg) {
	char buf[MBUFSIZE];
	compose(buf, sizeof(buf), "invalid!!", file, line, msg, 0);
	halt(buf);
}

void TJErrorPanic(const char* file, int line, const char* msg, int error) {
	char buf[MBUFSIZE];
	compose(buf, sizeof(buf), "error!!", file, line, msg, error);
	halt(buf);
}

void TJPrintWarning(const char* file, int line, const char* msg) {
	char buf[MBUFSIZE];
	compose(buf, sizeof(buf), "warning", file, line, msg, 0);
	writeMessage(buf);
}

/**
 * O݃G[
 */
void TJLogFailWarning(const char* file, int line, const char* msg) {
	char buf[MBUFSIZE];
	int error = TJGetLastError();
	compose(buf, sizeof(buf), "logging error!!", file, line, msg, error);
	fputs(buf, stderr);	/* ???????? o͐͗v ???????? */
	fputs("\n", stderr);
}
