/*-----------------------------------------------------------------------

      Copyright (c) 1998,1999 Eiwa System Management, Inc

                      All Rights Reserved

  This file contains proprietary source code and other proprietary
  material owned by Eiwa System Management, Inc. Published only in
  limited copyright sence.

------------------------------------------------------------------------*/
/*
 * $Id: TJThread.cc,v 1.26 1999/07/28 06:02:41 hiranabe Exp $
 */
#include "TJThread.h"

#include "TJRunnable.h"
#include "TJPrimitive.h"
#include "TJSystemInternal.h"
#include "tjdebug.h"
#include "tjprimdebug.h"
#include <stdio.h>	// sprintf()
#include <string.h>
#include <limits.h>

#if defined(TJ_THREAD_DEBUG)
#	include "TJThreadDebugger.h"
#else
#	define TH_PUSHNAME(s)
#	define TH_POPNAME()
#endif

/**
 * XbhJvZ邽߂̃NX
 *
 * ̃CuɂXbhƂ
 * \begin{verbatim}
 *     TJThread ܂͂̃TuNX̃\bh run() ̎s
 *     TJRunnable ̎NX̃\bh run() ̎s
 * \end{verbatim}
 * ̂ꂩӖĂ.
 *
 * XbhsJn̂̓XbhIuWFNgɑ΂ start() s
 * ł. ̎sJnXbhIuWFNg run() ɐ䂪
 * n. ̎dl̓Xbh쐬ƎsJn𕪗łȂlCeBu
 * XbhCu( POSIX thread )łێȂ΂ȂȂD
 *
 *  run() ̕Aɓ exit(0) Ăяo, XbhI
 *  exit() sƍlĂ悢.
 *
 * ThreadJack ł́C̃Xbhs삳邱ƂłD
 * yclassLockɊւ钍ӎz
 *     TJThread::classLock ̖ړI̓Xbh̍쐬EуXbh
 *     IuWFNg̒T(ƂcurrentThread()̎s) threadTable 
 *     ی삷邱ƂɂD
 *     ̃bNlĂ鎞
 *           ̂Ȃ郍bNlȂ
 *           Ȃ҂킹Ȃ
 *     ̂Q_炷邱ƁD̑O񂪂Ɖ͂eXgȂD
 *
 * \begin{verbatim}
 *     邽߂ɏ̘b JOINABLE Xbh join()
 *     Xbh̃ReLXgŃXbȟns JOINABLE 
 *     ͂ȂXbh exit() JgXbhŃXbȟn
 *     s.
 *      JOINABLE Xbhɑ΂ join() sȂΊǗ
 *     Xbh̖S삪܂ł݂[NN.
 *     t JOINABLE ł͂ȂXbh join() ꍇɂ͕sAN
 *     ZX̊댯.
 * \end{verbatim}
 *
 */


#define ClassLock()   classLock->lock()
#define ClassUnLock() classLock->unlock()


//#define INT_MAX  ((int)(~((unsigned)1 << (8 * (int)sizeof(int)) - 1)))
const int TJThread::JOINDETACHED = INT_MAX;
const int TJThread::JOINTIMEOUT = (INT_MAX-1);
const TJThread::ID TJThread::None = (TJThread::ID)(-1);
const int TJThread::MIN_PRIORITY = 1;
const int TJThread::MAX_PRIORITY = 10;
const unsigned TJThread::DEFSTACKSIZE = (16 * 1024);
const unsigned TJThread::NOTHREAD = UINT_MAX;

TJThread* TJThread::mainThread = 0;
bool TJThread::mainPurged = false;
TJThread::Assoc* TJThread::threadTable = 0;
TJMutex* TJThread::classLock = 0;
unsigned TJThread::serialno = 0;

static const char DEFMAINTHREAD[] = "Main";
static const char DEFFREETHREAD[] = "Free";


/**
 * VXeɓnXbh̃Gg|Cg
 *
 * Xbh{̂ C++\bhł͂Ȃ C֐`łȂ΂ȂȂD
 * P TJThread ̃|C^nD
 */
/*extern "C"*/
static void TJThreadbody(void* arg) {
	TJThread* thread = (TJThread*)arg;

#if defined(TJ_THREAD_DEBUG)
	/**
	 * TJThreadDebugger ̓fobOΏۂ̃Xbh̃ReLXg
	 * 쐬Ȃ΂ȂȂ.
	 * Xbh̃RXgN^͐ẽXbhł邱Ƃɒ.
	 */
	TJThreadDebugger::add(thread->getName());
#endif
	/**
	 * e start() ҂킹.
	 */
	int error = ::TJThreadWaitforStart(thread->getThreadID());
	if (error != 0)
		TJSystem_Error("TJThreadbody", error);
	/**
	 * tO running ̃Zbg start() ̒ōs.
	 * NA exit() ̒ōs.
	 * run() ̑OŃZbgƃNAł΂킩₷
	 * cOȂ炱̓NX̃XR[vOł.
	 */
	thread->run();
	thread->exit(0);
}

void TJThread::run() {
	if (target != 0)
		target->run();
}

/**
 * NXƂ̏̏
 *
 * CX^X쐬܂̓NX\bhďoɈxs
 * Ȃ΂ȂȂD
 *
 * ylzŏ̃Xbĥ main Xbĥ͂łD
 *         ŏ̃XbhOɃNX\bhďoꂽ
 *         ΁CĂяô main Xbĥ͂łD
 *         āC̃\bh͕K main XbhɂĎs
 *         ͂łD{? ^_^;)
 */
void TJThread::init() {
	if (threadTable == 0) {
		threadTable = new Assoc;
		mem_assert(threadTable);
		classLock = new TJMutex;
		mem_assert(classLock);
		int error = ::TJThreadInit();
		if (error != 0)
			TJSystem_Error("TJThread::init", error);
#if defined(TJ_THREAD_DEBUG)
		classLock->suppressPause();
#endif
		TJTRACE_INIT();
	}
}

/**
 * main XbhIuWFNg쐬D
 * ԍŏ̃Xbh͍쐬ς݂̂ߓʂȃRXgN^KvƂD
 * ̃RXgN^ private łD
 */
TJThread::TJThread(Kind k, TJThread::ID tid)
		: threadMutex(), joinSync(&threadMutex),
		  kind(k), target(0) {

	require(classLock->isLocked());
	assert(k == MAIN);

	parent = 0;
	joinptr = 0;
	exitcode = 0;
	anydata = 0;
	running = true;

	name = new char[::strlen(DEFMAINTHREAD)+1];
	mem_assert(name);
	::strcpy(name, DEFMAINTHREAD);
	threadID = tid;

	threadTable->put(threadID, this);

#if defined(TJ_THREAD_DEBUG)
	TJThreadDebugger::addMain(name);
#endif

#if defined(DEBUG) && defined(THMEMDEBUG)
	printf("@THMEM@ %s:%d: TJThread::TJThread() this=%x\n",
			name, (int)threadID, (unsigned)this);
#endif
}

/**
 * XbhIuWFNg̏
 */
void TJThread::construct(const char* name, TJThread* parent, ID curtid) {

	require(classLock->isLocked());

	if (parent == 0) {
		parent = mainThread = new TJThread(MAIN, curtid);
		mem_assert(parent);
	}
	this->parent = parent;
	this->anydata = 0;
	this->joinptr = 0;

	this->exitcode = 0;
	this->running = false;

	char defnamebuf[16];
	if (name == 0 || name[0] == '\0') {
		/*
		 * Xbhȗ̃ftHgD
		 */
		unsigned sno = ++serialno;	// must be class-locked
		::sprintf(defnamebuf, "t%u", sno);
		name = defnamebuf;
	}
	this->name = new char[::strlen(name)+1];
	mem_assert(this->name);
	::strcpy(this->name, name);

	this->threadID = None;
}

/**
 * JgXbh̃IuWFNgTuNX̃IuWFNgɕύX.
 *
 * ʏ̓TuNX̃IuWFNg쐬邽, ̃\bh͕Kvł͂Ȃ.
 * ̃\bh main XbhIuWFNgŕKvƂȂꍇ.
 */
void TJThread::exchange(TJThread* newthread) {

	TJThread* curthread = currentThread();
	if (curthread != mainThread)
		TJSystem_invalid("Must be the Main Thread.");

	ClassLock();

	newthread->parent = curthread->parent;
	newthread->anydata = curthread->anydata;
	newthread->joinptr = curthread->joinptr;
	newthread->exitcode = curthread->exitcode;
	newthread->running = curthread->running;
	newthread->setName(curthread->name);

	TJThread* original =
		threadTable->exchange(curthread->threadID, newthread);
	if (original == mainThread)
		mainThread = newthread;
	ClassUnLock();

	assert(original != 0);
	original->running = false;	// ~TJThread()
	delete original;
}

/**
 * VXeˑ̃Xbh쐬
 * ftHgł̓v~eBũXbh쐬Ăяo, Kvɉ
 * TuNXŃI[oCh\.
 */
int TJThread::create(void (*entrypoint)(void*), void* arg,
		unsigned stacksize, void* /*info*/, ID* ptid) {
	return ::TJThreadMake(name, stacksize, entrypoint, arg, ptid);
}

/**
 * Xbh̍쐬
 */
void TJThread::newContext(unsigned stacksize, void* info) {

	require(classLock->isLocked());

	int error = create(TJThreadbody, this, stacksize, info, &threadID);
	if (error != 0) {
		ClassUnLock();
		TJSystem_Error("TJThread::newContext", error);
	}

	/**
	 * XbhDx̓CX^XϐVXe@\̂܂
	 * g. قƂǂ̃VXeł͗Dx͎IɌp邾낤
	 * ł͂ȂVXeꍇɔĂŃVXeˑ
	 * z.
	 */
	int ppri = parent->getPriority();
	if (getPriority() != ppri)
		setPriority(ppri);

	threadTable->put(threadID, this);

#if defined(DEBUG) && defined(THMEMDEBUG)
	printf("@THMEM@ %s:%d: TJThread::TJThread() this=%x\n",
			name, (int)threadID, (unsigned)this);
#endif
}

/**
 * Xbh̍쐬
 */
void TJThread::createThread(unsigned stacksize, void* info) {
	ClassLock();
	newContext(stacksize, info);
	ClassUnLock();
}

/**
 * IȃRXgN^
 */
void TJThread::add(const char* name, unsigned stacksize) {
	if (threadTable == 0)
		init();

	ID curtid = currentThreadID();
	TJThread* parent = getThread(curtid);	// ŏ 0 Ȃ

	ClassLock();

	construct(name, parent, curtid);
	if (stacksize != NOTHREAD)
		newContext(stacksize, 0);

	ClassUnLock();
}

/**
 * IȃfXgN^
 */
void TJThread::remove() {
	if (classLock == 0)
		return;
	ClassLock();
	threadTable->remove(this->threadID);
	int count = threadTable->count;
#if defined(TJ_THREAD_DEBUG)
	/**
	 * قƂǂ̏ꍇ exit() ̒ō폜Ă.
	 */
	TJThreadDebugger::remove(this->threadID);
#endif
	if (this == mainThread) {
		mainThread = 0;
		mainPurged = true;
	}
	ClassUnLock();

	if (count <= 0) {
		delete threadTable;
		threadTable = 0;
		delete classLock;
		classLock = 0;
		TJTRACE_TERMINATE();
	}
#if defined(DEBUG) && defined(THMEMDEBUG)
	printf("@THMEM@ %s:%d: TJThread::~TJThread() this=%x nthreads=%d\n",
			name, (int)threadID, (unsigned)this, count);
#endif
	delete [] name;
}

/**
 * XbhʎqXbhIuWFNg߂D
 */
TJThread* TJThread::getThread(TJThread::ID tid) {
	if (threadTable == 0)
		init();
	ClassLock();
	TJThread* thread = threadTable->get(tid);
	if (thread == 0 && !mainPurged && mainThread == 0) {
		thread = new TJThread(MAIN, tid);	// main thread ?
		mem_assert(thread);
		mainThread = thread;
	}
	ClassUnLock();
	return thread;
}

/**
 * XbhԂ.
 */
int TJThread::activeCount() {
	ClassLock();
	int count = threadTable->count;
	ClassUnLock();
	return count;
}

/**
 * ׂẴXbh̃|C^w vector ɃZbg.
 */
//int TJThread::enumerate(vector<TJThread*>& vec) {
//	ClassLock();
//	int count = threadTable->count;
//	for (int n = 0; n < count; n++)
//		vec.push_back(threadTable->pointer[n].value);
//	ClassUnLock();
//	return count;
//}

/**
 * run() ̒sȂ true ԂD
 */
bool TJThread::isAlive() const {
	return running;		// atomic operation
}

/* ------------------------------------------------------------------------ */

TJThread::TJThread(unsigned stacksize /*= DEFSTACKSIZE*/)
		: threadMutex(), joinSync(&threadMutex),
		  kind(JOINABLE), target(0) {
	add("", stacksize);
}

TJThread::TJThread(const char* name, unsigned stacksize /*= DEFSTACKSIZE*/)
		: threadMutex(), joinSync(&threadMutex),
		  kind(JOINABLE), target(0) {
	add(name, stacksize);
}

TJThread::TJThread(TJRunnable* t, unsigned stacksize /*= DEFSTACKSIZE*/)
		: threadMutex(), joinSync(&threadMutex),
		  kind(JOINABLE), target(t) {
	add("", stacksize);
}

TJThread::TJThread(TJRunnable* t, const char* name,
		unsigned stacksize /*= DEFSTACKSIZE*/)
		: threadMutex(), joinSync(&threadMutex),
		  kind(JOINABLE), target(t) {
	add(name, stacksize);
}

TJThread::~TJThread() {
	/**
	 * Xbh̃TuNX public ȃfXgN^邩Ȃ.
	 */
	if (!running) {
		remove();
		return;
	}
	TJSystem_invalid("Please do not delete TJThread.");
}

/* ------------------------------------------------------------------------ */

void TJThread::setPriority(int pri) {
	if (pri < MIN_PRIORITY || pri > MAX_PRIORITY)
		TJSystem_invalid("pri < MIN_PRIORITY || pri > MAX_PRIORITY");
	int error = ::TJThreadSetPriority(threadID, pri);
	if (error != 0)
		TJSystem_Error("TJThread::setPriority", error);
}

int TJThread::getPriority() const {
	int npri = 0;
	int error = ::TJThreadGetPriority(threadID, &npri);
	if (error != 0)
		TJSystem_Error("TJThread::getPriority", error);
	return npri;
}

void TJThread::start() {
	running = true;
	int error = ::TJThreadStartup(threadID);
	if (error != 0)
		TJSystem_Error("TJThread::start", error);
}

/**
 * JgXbhID
 *
 * JOINABLE XbhłȂ΃JgXbhŃXbhǗ
 * ̍폜ƃsD
 * JgXbhŌ̃Xbh̏ꍇ̓XbhI
 * ͂ȂvZXIƂȂ.
 *
 * Xbh̏IɂĂ exit()  join() ̓ TJThread IuWFNg
 * ̍폜ȂǁCȖ肪D
 * ̃eXgvO͂̖𔭌ĂD(Solaris)
 *
 *   yRpCz
 *       % g++ -c -g -Wall -D_REENTRANT -DTJ_THREAD_DEBUG -DDEBUG \
 *           -DTHMEMDEBUG TJThread.cc
 *       % g++ -c -g -Wall -D_REENTRANT -DTJ_THREAD_DEBUG -DDEBUG Test.cc
 *       % g++ -o Test -g Test.o TJThread.o libsysfw.a -lthread -lnsl -lsocket
 *
 *   yӎz
 *       RpC -DTJ_THREAD_DEBUG -DTHMEMDEBUG IvV
 *       w肷邱ƁD
 *       ̌x͖Ă悢D
 *             warning: implicit declaration of function `int sleep(...)'
 *       100 ȏ̈Ŏs join()  exit() D
 *       ~NȃbNxŃeXgꍇ
 *             setenv THPORT 32111
 *             Test
 *       ܂
 *             env THPORT=32111 Test
 *       Ƃ邱ƁD|[gԍ͋̉\ȂΉł悢D
 *
 *   #include <stdio.h>
 *   #include <stdlib.h>
 *   #include <string>
 *   #include "pfmt.h"
 *
 *   class Child : public TJThread {
 *       private:
 *           int code;
 *       public:
 *           Child(const char* name, int c) : TJThread(name) { code = c; }
 *           ~Child() { }
 *           void run() {
 *                   printf("@@@ %s: Child started\n", TJThreadName());
 *                   if (code != 0)
 *                           TJThread::exit(code);
 *           }
 *   };
 *
 *   int main(int argc, char** argv) {
 *           const char* e = getenv("THPORT");
 *           int port = (e != 0)? atoi(e) : 0;
 *           SETDEBUGPORT(port);
 *           int c = (argc >= 2)? atoi(argv[1]) : 0;
 *           Child* child = new Child("child", c);
 *           child->start();
 *           if (c >= 100)
 *                   TJThread::sleep(1500);
 *           int code = child->join();
 *           delete child;
 *           printf("*** %s: delete done. code=%d\n", TJThreadName(), code);
 *           TJThread::sleep(3000);
 *           TJThread::exit(0);
 *           return 0;
 *   }
 */
void TJThread::exit(int code) {
	TH_PUSHNAME("TJThread::exit");	// for debug

	TJThread* exitthread = currentThread();

	exitthread->threadMutex.lock();

	/**
	 * JgXbh JOINABLE łȂꍇ exitcode, running ̍XV
	 *  signal() ̎ssȂƂI蓾D
	 */
	exitthread->exitcode = code;
	exitthread->running = false;

	TJThread* waitthread = exitthread->joinptr;
	if (waitthread != 0)
		exitthread->joinSync.signal();

#if defined(DEBUG) && defined(THMEMDEBUG)
	if (waitthread != 0)
		printf("@THMEM@ %s:%d: TJThread::exit() waitThread=%s\n",
			exitthread->name, (int)exitthread->threadID,
			waitthread->name);
	else
		printf("@THMEM@ %s:%d: TJThread::exit() waitThread=?\n",
			exitthread->name, (int)exitthread->threadID);
#endif

#if defined(TJ_THREAD_DEBUG)
	TJThreadDebugger::remove(exitthread->threadID);
#endif

	bool selfDelete = exitthread->kind != JOINABLE;

	/**
	 * ******************* x *************************************
	 * ̃bN TJThread ̃IuWFNg delete 
	 * \l邱ƁD
	 * **************************************************************
	 */
	exitthread->threadMutex.unlock();

#if defined(DEBUG) && defined(THMEMDEBUG)
	/**
	 * ̈ӂ TJThread IuWFNg delete s core dump
	 * ȂƂmF邽߂ sleep().
	 * TJThread::sleep() ̓IuWFNgKvƂ邽ߎgȂƁD
	 */
	//::sleep(1);
#endif

	/**
	 * JOINABLE XbhłȂ΃JgXbhŃXg̍폜
	 * sD
	 */
	if (selfDelete)
		delete exitthread;

	/**
	 * TJThreadTerminate() ̒ŃJgXbh͏I邽
	 * ̊֐琧䂪߂邱Ƃ͂ȂD
	 * JgXbhŌ̃Xbh̏ꍇ̓XbhI
	 * ł͂ȂvZXIƂȂ.
	 */
	if (threadTable == 0)	// see TJThread::remove()
		::TJSystemExit(code);	// vZXI
	else
		::TJThreadTerminate(code);	// XbhI
}

/**
 * XbhI҂킹D
 *
 * ڐÂ Solaris ̏ꍇ thr_join() gȂD
 *
 * JOINABLE ł͂ȂXbhɑ΂ join() sꍇ̓
 * słD
 *  exit() sĂ t->join()  core dump \D
 * ^ǂ join() ɂǂƂĂ exitthread->threadMutex ̎s
 * D
 * joinable ł͂ȂXbhɑ΂ join() sꍇ̓
 * ͕sł. G[߂l JOINDETACHED ͑肪IOȂ
 * Ԃ邪łȂΕsANZXƂȂ.
 * joinable ł͂ȂXbh͏IɃXbhIuWFNg̈
 * . t JOINABLE Xbh join() ŃIuWFNg
 * ̈.
 */
int TJThread::join(long milliseconds) {
	TH_PUSHNAME("TJThread::join");	// for debug

	TJThread* exitthread = this;
	TJThread* waitthread = currentThread();

	exitthread->threadMutex.lock();

	if (exitthread->kind != JOINABLE) {
		exitthread->threadMutex.unlock();
		assert(false);
		TH_POPNAME();	// for debug
		return JOINDETACHED;
	}

#if defined(DEBUG) && defined(THMEMDEBUG)
	printf("@THMEM@ %s:%d: TJThread::join() waitThread=%s\n",
			exitthread->name, (int)exitthread->threadID,
			waitthread->name);
#endif

	while (exitthread->running) {
		exitthread->joinptr = waitthread;
		if (milliseconds <= (long)0) {
			exitthread->joinSync.wait();
		} else {
			if (!exitthread->joinSync.timedwait(milliseconds)) {
				exitthread->threadMutex.unlock();
				TH_POPNAME();	// for debug
				return JOINTIMEOUT;
			}
		}
	}

	int exitcode = exitthread->exitcode;

	exitthread->threadMutex.unlock();

	/**
	 * [Uɂ delete sȂdlƂ.
	 * Ȃ TJThread::~TJThread()  proteced ɂĂ邽
	 * X^bNɒuƃRpCG[ƂȂ͂.
	 */
	delete exitthread;

	TH_POPNAME();	// for debug
	return exitcode;
}

int TJThread::join() {
	return join((long)0);
}

void TJThread::sleep(long milliseconds) {
	TH_PUSHNAME("TJThread::sleep");	// for debug

	if (threadTable == 0)
		init();

	if (milliseconds <= (long)0) {
		::TJThreadYield();
	} else {
		int error = ::TJThreadSleep(milliseconds);
		if (error != 0)
			TJSystem_Error("TJThread::sleep", error);
	}

	TH_POPNAME();	// for debug
}

void TJThread::yield() {
	if (threadTable == 0)
		init();
	::TJThreadYield();
}

TJThread::ID TJThread::currentThreadID() {
	if (threadTable == 0)
		init();
	return (ID)::TJThreadSelf();
}

TJThread* TJThread::currentThread() {
	if (threadTable == 0)
		init();
	ID tid = (ID)::TJThreadSelf();
	TJThread* thread = getThread(tid);
#if defined(DEBUG) && defined(THMEMDEBUG)
	if (thread == 0)
		printf("@THMEM@ %d: currentThread() == 0\n", (int)tid);
#endif
	assert(thread != 0);
	return thread;
}

const char* TJThread::currentThreadName() {
	TJThread* curthread = currentThread();
	return (curthread != 0)? curthread->name : " ? ";
}

bool TJThread::isJoinable() const {
	threadMutex.lock();
	bool j = kind == JOINABLE;
	threadMutex.unlock();
	return j;
}

void TJThread::setJoinable(bool on) {
	if (this == mainThread)
		return;
	threadMutex.lock();
	kind = on? JOINABLE : FREE;
	threadMutex.unlock();
}

void TJThread::putData(void* anydata) {
	threadMutex.lock();
	this->anydata = anydata;
	threadMutex.unlock();
}

void* TJThread::getData() const {
	threadMutex.lock();
	void* anydata = this->anydata;
	threadMutex.unlock();
	return anydata;
}

void TJThread::setName(const char* name) {
	threadMutex.lock();
	delete [] this->name;
	this->name = new char[::strlen(name)+1];
	mem_assert(this->name);
	::strcpy(this->name, name);
	threadMutex.unlock();
}

const char* TJThread::getName() const {
	return name;
}

TJThread::ID TJThread::getThreadID() const {
	return threadID;
}

/* -------------------------- TJThread::Assoc ----------------------------- */
/**
 * XbhʎqXbhIuWFNgւ̍ϊ
 *
 * put(), remove() ͌ get() ̓XbhĂrI
 * DpɂɃXbhE폜sprł map ɒuȂǂ
 * ΍􂪕KvȂD
 */

TJThread::Assoc::Assoc(int inisize /*= 16*/) {
	psize = inisize;
	count = 0;
	pointer = new Pair[psize];
	mem_assert(pointer);
}

TJThread::Assoc::~Assoc() {
	if (pointer != 0) {
		delete [] pointer;
		pointer = 0;
	}
}

void TJThread::Assoc::put(TJThread::ID key, TJThread* value) {
	require(classLock->isLocked());
	if (count >= psize) {
		int newsize = psize * 2;
		Pair* newptr = new Pair[newsize];
		mem_assert(newptr);
		::memcpy(newptr, pointer, sizeof(Pair) * psize);
		delete [] pointer;
		pointer = newptr;
		psize = newsize;
	}
	int index = 0;
	for ( ; index < count; index++)
		if (key < pointer[index].key)
			break;
	for (int n = count; n > index; n--) {
		pointer[n].key = pointer[n-1].key;
		pointer[n].value = pointer[n-1].value;
	}
	pointer[index].key = key;
	pointer[index].value = value;
	count++;
}

void TJThread::Assoc::remove(TJThread::ID key) {
	require(classLock->isLocked());
	int index = getIndex(key);
	if (index < 0)
		return;
	for (int max = count - 1, n = index; n < max; n++) {
		pointer[n].key = pointer[n+1].key;
		pointer[n].value = pointer[n+1].value;
	}
	count--;
}

TJThread* TJThread::Assoc::get(TJThread::ID key) const {
	require(classLock->isLocked());
	int index = getIndex(key);
	return (index >= 0)? pointer[index].value : 0;
}

/**
 * main XbhTuNX̃IuWFNgɌ.
 */
TJThread* TJThread::Assoc::exchange(ID key, TJThread* newthread) {
	require(classLock->isLocked());
	int index = getIndex(key);
	if (index < 0)
		return 0;
	TJThread* original = pointer[index].value;
	ID tmp = newthread->threadID;
	newthread->threadID = original->threadID;
	original->threadID = tmp;
	pointer[index].value = newthread;
	return original;
}

int TJThread::Assoc::getIndex(TJThread::ID key) const {
	int low = 0;
	int high = count - 1;

	while (low <= high) {
		int mid = (low + high) / 2;
		if (key < pointer[mid].key)
			high = mid - 1;
		else if (key > pointer[mid].key)
			low = mid + 1;
		else
			return mid;
	}
	return -1;
}
