/*-----------------------------------------------------------------------

      Copyright (c) 1998,1999 Eiwa System Management, Inc

                      All Rights Reserved

  This file contains proprietary source code and other proprietary
  material owned by Eiwa System Management, Inc. Published only in
  limited copyright sence.

------------------------------------------------------------------------*/
/*
 * $Id: TJUtility.cc,v 1.3 1999/04/12 11:09:01 hiranabe Exp $
 */
#include "TJUtility.h"
#include "TJPrimitive.h"
#include <stdio.h>	// sprintf()


string TJUtility::separator() {
	return string(::TJFileSeparator());
}

string TJUtility::sysdepFilename(const string& filename) {
	string syssep = ::TJPathSeparator();
	string stdsep = "/";
	string name = filename;
	if (syssep != stdsep) {
		string::size_type pos = 0;
		while ((pos = filename.find(stdsep, pos)) != string::npos) {
			name.replace(pos, stdsep.size(), syssep);
			pos++;
		}
	}
	return name;
}

string TJUtility::toHexString(int byte) {
	static char hex[] = "0123456789abcdef";
	string s = string("") + (char)hex[byte >> 4 & 0x0f];
	s += hex[byte & 0x0f];
	return s;
}

string TJUtility::toHexString(const char* data, int count,
		int byteperline /*=16*/) {
	string s = "";
	string sep = separator();
	for (int n = 0; n < count; n++) {
		if (n > 0)
			s += (n % byteperline != 0)? string(" ") : sep;
		s += toHexString(*data++);
	}
	if (count > byteperline)
		s += sep;
	return s;
}

string TJUtility::toPrintable(const char* data, int count) {
	string s = "";
	for (const char* end = data + count; data < end; data++) {
		char c = (0x20 <= *data && *data <= 0x7e)? *data : '.';
		s += c;
	}
	return s;
}

string TJUtility::toHexDumpFormat(unsigned offset, const char* data, int count,
		int byteperline /*=16*/) {
	string s = "";
	string sep = separator();
	char num[16];
	while (count >= byteperline) {
		::sprintf(num, "%04x", offset);
		s += string(num) + "  ";
		s += toHexString(data, byteperline, byteperline) + "    ";
		s += toPrintable(data, byteperline) + sep;
		offset += byteperline;
		data += byteperline;
		count -= byteperline;
	}
	if (count > 0) {
		::sprintf(num, "%04x", offset);
		s += string(num) + "  ";
		s += toHexString(data, count, byteperline) + "    ";
		for (int n = byteperline - count; n > 0; n--)
			s += "   ";
		s += toPrintable(data, count) + sep;
	}
	return s;
}

string TJUtility::toBinString(int byte) {
	string s = "";
	for (int n = 0; n < 8; n++) {
		s += ((byte & 0x80) == 0)? '0' : '1';
		byte <<= 1;
	}
	return s;
}

string TJUtility::toBinString(const char* data, int count,
		int byteperline /*=8*/) {
	string s = "";
	string sep = separator();
	for (int n = 0; n < count; n++) {
		if (n > 0)
			s += (n % byteperline != 0)? string(" ") : sep;
		s += toBinString(*data++);
	}
	if (count > byteperline)
		s += sep;
	return s;
}
