/*-----------------------------------------------------------------------

      Copyright (c) 1998,1999 Eiwa System Management, Inc

                      All Rights Reserved                         

  This file contains proprietary source code and other proprietary
  material owned by Eiwa System Management, Inc. Published only in
  limited copyright sence.

------------------------------------------------------------------------*/
/*
 * $Id: TJWMsgQ.c,v 1.1 1999/05/21 07:20:41 tokui Exp $
 */
/**
 * Windows ł̃v~eBu[`
 */

#include <windows.h>
#include <assert.h>
#include <errno.h>
#include "TJPrimitive.h"


/*****************************************************************************
 *                   bZ[WL[֘Ãv~eBu
 *****************************************************************************/

typedef struct {
	void* mutex;
	void* cond;
	void* revcond;
	int   iswaiting;
	int   isrevwaiting;
	void* buffer;
	int   maxmsgs;
	int   maxmsglen;
	int*  lenbuf;
	int   rptr;
	int   count;
	int   broken;
	int   deletable;
	int   revdeletable;
} mqinfo_t;

/**
 * bZ[WL[𐶐.
 */
int TJQueueCreate(const char* name, int maxmsgs, int maxmsglen, void** pmq) {
	int error;
	mqinfo_t* mq = (mqinfo_t*)malloc(sizeof(mqinfo_t));
	if (mq == (mqinfo_t*)0)
		return ENOMEM;

	mq->buffer = malloc(maxmsgs * maxmsglen);
	if (mq->buffer == (void*)0)
		return ENOMEM;

	mq->lenbuf = (int*)malloc(maxmsgs * sizeof(int));
	if (mq->lenbuf == (int*)0)
		return ENOMEM;

	error = TJMutexCreate(&mq->mutex);
	if (error)
		return error;
	error = TJSyncCreate(&mq->cond);
	if (error)
		return error;
	error = TJSyncCreate(&mq->revcond);
	if (error)
		return error;

	mq->iswaiting = mq->isrevwaiting = 0;
	mq->maxmsgs = maxmsgs;
	mq->maxmsglen = maxmsglen;
	mq->rptr = 0;
	mq->count = 0;
	mq->broken = mq->deletable = mq->revdeletable = 0;
	*pmq = (void*)mq;
	return 0;
}

/**
 * bZ[WL[.
 *
 * j󂳂悤ƂĂL[ő҂ĂXbh̏
 */
void TJQueueDelete(void* mque, const char* name) {
	mqinfo_t* mq = (mqinfo_t*)mque;

	mq->broken = 1;
	if (mq->iswaiting > 0) {
		TJSyncNotifyAll(mq->cond);
		while (!mq->deletable)
			TJThreadYield();
	}
	if (mq->isrevwaiting > 0) {
		TJSyncNotifyAll(mq->revcond);
		while (!mq->revdeletable)
			TJThreadYield();
	}

	TJMutexDelete(mq->mutex);
	TJSyncDelete(mq->cond);
	TJSyncDelete(mq->revcond);

	free(mq->buffer);
	free(mq->lenbuf);
	free((void*)mq);
}

/**
 * bZ[W𑗐M.
 *
 * dl킹邽߂ɃL[̃TCY͌ŒƂCL[ɋ󂫂Ȃꍇ
 * ̓ubND
 */
int TJQueueSend(void* mque, const void* msg, int length) {
	int error;
	char* destp;
	int index;
	mqinfo_t* mq = (mqinfo_t*)mque;

	assert(length <= mq->maxmsglen);

	error = TJMutexLock(mq->mutex);
	if (error)
		return error;

	while (mq->count >= mq->maxmsgs) {
		mq->isrevwaiting++;
		error = TJSyncWait(mq->revcond, mq->mutex);
		if (mq->broken) {
			mq->revdeletable = 1;
			return -1;
		}
		if (error && error != EINTR)
			return error;
	}

	index = (mq->rptr + mq->count) % mq->maxmsgs;
	destp = (char*)mq->buffer + index * mq->maxmsglen;
	memcpy((void*)destp, msg, length);
	mq->lenbuf[index] = length;

	mq->count++;

	if (mq->iswaiting > 0) {
		mq->iswaiting--;
		error = TJSyncNotify(mq->cond);
		if (error)
			return error;
	}

	error = TJMutexUnlock(mq->mutex);
	if (error)
		return error;
	return 0;
}

/**
 * bZ[WM.
 */
int TJQueueReceive(void* mque, void* msgbuf, int bufsize, int* plen,
		int milliseconds) {
	int length = -1;
	int error;
	char* srcp;
	mqinfo_t* mq = (mqinfo_t*)mque;

	error = TJMutexLock(mq->mutex);
	if (error)
		return error;

	if (milliseconds == 0 && mq->count <= 0)
		goto out;

	while (mq->count <= 0) {
		mq->iswaiting++;
		error = (milliseconds < 0)? TJSyncWait(mq->cond, mq->mutex) :
			TJSyncWaitTimeout(mq->cond, mq->mutex, milliseconds);
		if (mq->broken) {
			mq->deletable = 1;
			return -1;
		}
		if (error < 0) {	/* timeout */
			mq->iswaiting--;
			goto out;
		}
		if (error && error != EINTR)
			return error;
	}

	length = mq->lenbuf[mq->rptr];
	assert(bufsize >= length);
	srcp = (char*)mq->buffer + mq->rptr * mq->maxmsglen;
	memcpy(msgbuf, (const void*)srcp, length);

	mq->count--;
	mq->rptr++;
	if (mq->rptr >= mq->maxmsgs)
		mq->rptr = 0;

	if (mq->isrevwaiting > 0) {
		mq->isrevwaiting--;
		error = TJSyncNotify(mq->revcond);
		if (error)
			return error;
	}

out:
	error = TJMutexUnlock(mq->mutex);
	if (error)
		return error;
	*plen = length;
	return 0;
}
